@@ -69,6 +69,7 @@
 #include <netinet/sctp.h>
 #include <unistd.h>
 #include <string.h>
+#include <signal.h>
 #include "sctp_darn.h"
 
 char *TCID = __FILE__;
@@ -78,6 +79,24 @@ int TST_CNT = 0;
 #define GEN_DATA_FIRST	0x21
 #define GEN_DATA_LAST	0x7e
 
+/* Display an IPv4 address in readable format.  */
+#define NIPQUAD(addr) \
+        ((unsigned char *)&addr)[0], \
+        ((unsigned char *)&addr)[1], \
+        ((unsigned char *)&addr)[2], \
+        ((unsigned char *)&addr)[3]
+
+/* Display an IPv6 address in readable format.  */
+#define NIP6(addr) \
+        ntohs((addr).s6_addr16[0]), \
+        ntohs((addr).s6_addr16[1]), \
+        ntohs((addr).s6_addr16[2]), \
+        ntohs((addr).s6_addr16[3]), \
+        ntohs((addr).s6_addr16[4]), \
+        ntohs((addr).s6_addr16[5]), \
+        ntohs((addr).s6_addr16[6]), \
+        ntohs((addr).s6_addr16[7])
+
 /* These are the global options.  */
 char *local_host = NULL;
 int local_port = 0;
@@ -88,6 +107,8 @@ struct sockaddr *bindx_add_addrs = NULL;
 int bindx_add_count = 0;
 struct sockaddr *bindx_rem_addrs = NULL;
 int bindx_rem_count = 0;
+struct sockaddr *connectx_addrs = NULL;
+int connectx_count = 0;
 int interactive_mode = 0;
 int poll_skn = 0;
 int nonblocking = 0;
@@ -103,6 +124,11 @@ sctp_assoc_t associd = 0;
 int echo = 0;
 char *interface = "eth0";
 int if_index = 0;
+sockaddr_storage_t remote_addr;
+sa_family_t ra_family;	/* What family is remote_addr? */
+int ra_len = 0;		/* How long is remote_addr? */
+void *ra_raw;		/* This is the addr part of remote_addr. */
+int new_connection = 1;
 
 enum inter_cmd_num {
 	INTER_SND = 0,
@@ -119,6 +145,11 @@ enum inter_cmd_num {
 	INTER_MAXSEG
 };
 
+enum shutdown_type {
+	SHUTDOWN_ABORT = 0,
+	SHUTDOWN_SHUTDOWN
+};
+
 struct inter_entry {
 	char *cmd;
 	int cmd_num;
@@ -154,17 +185,19 @@ static void snd_func(char *);
 static void sndbuf_func(char *, int, int, int);
 static void rcvbuf_func(char *, int, int, int);
 static struct sockaddr *get_bindx_addr(char *, int *);
-static int bindx_func(char *, int, struct sockaddr *, int, int);
+static int bindx_func(char *, int, struct sockaddr *, int, int, int);
+static int connectx_func(char *, int, struct sockaddr *, int);
 static void  primary_func(char *, int, char *, int);
 static void  peer_primary_func(char *, int, char *, int);
-static int abort_func(char *, int);
 static int nodelay_func(char *, int, int val, int set);
 static int maxseg_func(char *, int, int val, int set);
-static int shutdown_func(char *, int);
+static int shutdown_func(char *argv0, int *skp, int shutdown_type);
 static int test_sk_for_assoc(int sk, sctp_assoc_t assoc_id);
 static char * gen_message(int);
 static sctp_assoc_t test_recv_assoc_change(int);
 static sctp_assoc_t test_verify_assoc_change(struct msghdr *);
+void print_addr_buf(void * laddrs, int n_laddrs);
+int print_sockaddr(struct sockaddr *sa_addr);
 
 int
 main(int argc, char *argv[]) {
@@ -172,6 +205,8 @@ main(int argc, char *argv[]) {
 	int error = 0;
 	int i;
 
+	signal(SIGPIPE, SIG_IGN);
+
 	parse_arguments(argc, argv);
 
 	switch(command) {
@@ -186,7 +221,7 @@ main(int argc, char *argv[]) {
 		break;
 	case COMMAND_SEND:
 		sk = build_endpoint(argv[0], local_port);
-		error = command_send(argv[0], sk);
+		error = command_send(argv[0], &sk);
 		break;
 	case COMMAND_POLL:
 		if (use_poll) {
@@ -249,12 +284,13 @@ parse_arguments(int argc, char *argv[]) {
 		{"use-poll",	0, 0, 20},
 		{"echo",        0, 0, 'e'},
 		{"interface",   optional_argument, 0, 5,},
+		{"connectx",    1, 0, 17},
 		{0,		0, 0, 0}
 	};
 
 	/* Parse the arguments.  */
 	while (1) {
-		c = getopt_long (argc, argv, "B:H:IP:b:h:i:p:lm:nstz:e",
+		c = getopt_long (argc, argv, "B:H:IP:b:h:i:p:lm:nstz:ec:",
 				 long_options, &option_index);
 		if (c == -1)
 			break;
@@ -351,6 +387,22 @@ parse_arguments(int argc, char *argv[]) {
 			}
 			bindx_rem_addrs = tmp_addrs;
 			break;
+		case 17:		/* connectx */
+		case 'c':
+			tmp_addrs =
+				append_addr(optarg, connectx_addrs,
+					    &connectx_count);
+			if (NULL == tmp_addrs) {
+				/* We have no memory, so keep fprintf()
+				 * from trying to allocate more.
+				 */
+				fprintf(stderr, "No memory to add ");
+				fprintf(stderr, optarg);
+				fprintf(stderr, "\n");
+				exit(2);
+			}
+			connectx_addrs = tmp_addrs;
+			break;
 		case 20:		/* use-poll */
 			use_poll = 1;
 			break;
@@ -387,7 +439,7 @@ parse_arguments(int argc, char *argv[]) {
 			break;
 		case '?':
 			usage(argv[0]);
-			break;
+			exit(1);
 
 		default:
 			printf ("%s: unrecognized option 0%c\n",
@@ -416,6 +468,20 @@ parse_arguments(int argc, char *argv[]) {
 		exit(1);
 	}
 
+	if (command == COMMAND_SEND && NULL == remote_host
+	    && connectx_count == 0) {
+		fprintf(stderr, "%s: You MUST provide a remote host for sending.\n",
+			argv[0]);
+		usage(argv[0]);
+		exit(1);
+	}
+
+	if (remote_host != NULL && connectx_count != 0) {
+		fprintf(stderr, "%s: You can not provide both -h and -c options.\n",
+			argv[0]);
+		usage(argv[0]);
+		exit(1);
+	}
 } /* parse_arguments() */
 
 /* Set up the local endpoint.  */
@@ -496,7 +562,8 @@ build_endpoint(char *argv0, int portnum)
 	/* Do we need to do bindx() to add any additional addresses? */
 	if (bindx_add_addrs) {
 		if (0 != bindx_func(argv0, retval, bindx_add_addrs,
-			bindx_add_count, SCTP_BINDX_ADD_ADDR)) {
+			bindx_add_count, SCTP_BINDX_ADD_ADDR, portnum)) {
+			fprintf(stderr, "bindx_func (add) failed.\n");
 			exit(1);
 		}
 	} /* if (bindx_add_addrs) */
@@ -504,7 +571,8 @@ build_endpoint(char *argv0, int portnum)
 	/* Do we need to do bindx() to remove any bound addresses? */
 	if (bindx_rem_addrs) {
 		if (0 != bindx_func(argv0, retval, bindx_rem_addrs,
-			bindx_rem_count, SCTP_BINDX_REM_ADDR)) {
+			bindx_rem_count, SCTP_BINDX_REM_ADDR, portnum)) {
+			fprintf(stderr, "bindx_func (remove) failed.\n");
 			exit(1);
 		}
 	} /* if (bindx_rem_addrs) */
@@ -591,9 +659,19 @@ command_listen(char *argv0, int sk)
 	while (!done) {
 		if (interactive_mode) {
 			/* Read from the user.  */
-			printf("%s:%d-%s:%d Interactive mode> ",
-				local_host, local_port, remote_host,
-				remote_port);
+			if (remote_host) {
+				printf("%s:%d-%s:%d Interactive mode> ",
+					local_host, local_port, remote_host,
+					remote_port);
+			} else {
+				printf("%s:%d-", local_host, local_port);
+				if (associd) {
+					print_sockaddr(&remote_addr.sa);
+				} else {
+					printf("?:%d", remote_port);
+				}
+				printf(" Interactive mode> ");
+			}
 			fflush(stdout);
 			if (NULL == fgets(message, REALLY_BIG, stdin)) {
 				done = 1;
@@ -611,7 +689,7 @@ command_listen(char *argv0, int sk)
 		}
 
 		if (socket_type == SOCK_STREAM) {
-			int len = 0;
+			socklen_t len = 0;
 
 			if (!recvsk) {
 				if ((recvsk = accept(sk, NULL, &len)) < 0) {
@@ -687,7 +765,7 @@ command_listen(char *argv0, int sk)
 
 /* Read lines from stdin and send them to the socket.  */
 int
-command_send(char *argv0, int sk)
+command_send(char *argv0, int *skp)
 {
 	struct msghdr outmsg;
 	struct iovec iov;
@@ -695,76 +773,77 @@ command_send(char *argv0, int sk)
 	char message[REALLY_BIG];
 	struct hostent *hst;
 	int c;
-
-	sockaddr_storage_t remote_addr;
-	sa_family_t ra_family;	/* What family is remote_addr? */
-	int ra_len;		/* How long is remote_addr? */
-	void *ra_raw;		/* This is the addr part of remote_addr. */
-
+	struct sockaddr *addrs;
 	int msglen;
 	int error = 0;
+	int sk = *skp;
 
 	/* Set up the destination.  */
-	hst = gethostbyname(remote_host);
-	if (hst == NULL) {
-		hst = gethostbyname2(remote_host, AF_INET6);
-	}
+	if (remote_host != NULL) {
+		hst = gethostbyname(remote_host);
+		if (hst == NULL) {
+			hst = gethostbyname2(remote_host, AF_INET6);
+		}
 
-	if (hst == NULL || hst->h_length < 1) {
-		fprintf(stderr, "%s: bad hostname: %s\n",
-			argv0, remote_host);
-		exit(1);
-	}
+		if (hst == NULL || hst->h_length < 1) {
+			fprintf(stderr, "%s: bad hostname: %s\n",
+				argv0, remote_host);
+			exit(1);
+		}
 
-	ra_family = hst->h_addrtype;
-	switch (ra_family) {
-	case AF_INET:
-		ra_len = sizeof(remote_addr.v4);
-		ra_raw = &remote_addr.v4.sin_addr;
-		remote_addr.v4.sin_port = htons(remote_port);
-		remote_addr.v4.sin_family = AF_INET;
-		break;
-	case AF_INET6:
-		ra_len = sizeof(remote_addr.v6);
-		ra_raw = &remote_addr.v6.sin6_addr;
-		remote_addr.v6.sin6_port = htons(remote_port);
-		remote_addr.v6.sin6_family = AF_INET6;
-		remote_addr.v6.sin6_scope_id = if_index;
-		break;
-	default:
-		fprintf(stderr, "Invalid address type.\n");
-		exit(1);
-		break;
+		ra_family = hst->h_addrtype;
+		switch (ra_family) {
+		case AF_INET:
+			ra_len = sizeof(remote_addr.v4);
+			ra_raw = &remote_addr.v4.sin_addr;
+			remote_addr.v4.sin_port = htons(remote_port);
+			remote_addr.v4.sin_family = AF_INET;
+			break;
+		case AF_INET6:
+			ra_len = sizeof(remote_addr.v6);
+			ra_raw = &remote_addr.v6.sin6_addr;
+			remote_addr.v6.sin6_port = htons(remote_port);
+			remote_addr.v6.sin6_family = AF_INET6;
+			remote_addr.v6.sin6_scope_id = if_index;
+			break;
+		default:
+			fprintf(stderr, "Invalid address type.\n");
+			exit(1);
+			break;
+		}
+		memcpy(ra_raw, hst->h_addr_list[0], hst->h_length);
 	}
-	memcpy(ra_raw, hst->h_addr_list[0], hst->h_length);
-
 
 	/* Initialize the global value for interactive mode functions.  */
 	if (interactive_mode) {
 		inter_sk = sk;
 	}
 
-	/* Initialize the message struct we use to pass messages to
-	 * the remote socket.
-	 */
-	iov.iov_base = message;
-	iov.iov_len = REALLY_BIG;
-	outmsg.msg_iov = &iov;
-	outmsg.msg_iovlen = 1;
-	outmsg.msg_control = NULL;
-	outmsg.msg_controllen = 0;
-	outmsg.msg_name = &remote_addr;
-	outmsg.msg_namelen = ra_len;
-
 	printf("%s ready to send...\n", argv0);
 	while (!done) {
 		/* Read from the user.  */
-		if (interactive_mode) {
-			printf("%s:%d-%s:%d Interactive mode> ",
-				local_host, local_port, remote_host,
-				remote_port);
+		if (remote_host) {
+			if (interactive_mode) {
+				printf("%s:%d-%s:%d Interactive mode> ",
+					local_host, local_port, remote_host,
+					remote_port);
+			} else {
+				printf("%s:%d-%s:%d> ",
+				       local_host, local_port,
+				       remote_host, remote_port);
+			}
 		} else {
-			printf("%s:%d> ", remote_host, remote_port);
+			printf("%s:%d-", local_host, local_port);
+			if (associd) {
+				print_sockaddr(&remote_addr.sa);
+			} else {
+				printf("XXXXXX:%d", remote_port);
+			}
+			if (interactive_mode) {
+				printf(" Interactive mode> ");
+			} else {
+				printf("> ");
+			}
 		}
 		fflush(stdout);
 		if (NULL == fgets(message, REALLY_BIG, stdin)) {
@@ -777,7 +856,6 @@ command_send(char *argv0, int sk)
 			if (0 <= (c = parse_inter_commands(argv0, message,
 				1))) {
 				if (INTER_SND == c) {
-
 					iov.iov_base = inter_outbuf;
 					msglen = inter_outlen;
 					iov.iov_len = msglen;
@@ -803,10 +881,87 @@ command_send(char *argv0, int sk)
 		    (0 != test_sk_for_assoc(sk, associd))) {
 			associd = test_recv_assoc_change(sk);
 			printf("Old association gone, Starting a new one!\n");
+			new_connection = 1;
+		}
+
+		if (new_connection && connectx_count != 0) {
+			/* Do a sctp_connectx() to establish a connection. */
+			error = connectx_func(argv0, sk, connectx_addrs,
+					      connectx_count);
+			if (0 != error) {
+				if (error == -2) {
+					printf("Connection refused\n");
+					if (SOCK_SEQPACKET == socket_type) {
+						associd = test_recv_assoc_change(sk);
+					}
+					continue;
+				}
+				fprintf(stderr, "connectx failed.\n");
+				exit(1);
+			}
+			if (SOCK_SEQPACKET == socket_type) {
+				associd = test_recv_assoc_change(sk);
+			} else {
+				associd = 1;
+			}
+			int rc = sctp_getpaddrs(sk, associd, &addrs);
+			if (0 >= rc) {
+				if (rc == 0) {
+					fprintf(stderr, "sctp_getpaddrs failed, no peers.\n");
+				} else {
+					fprintf(stderr, "sctp_getpaddrs failed %s(%d).\n", strerror(errno), errno);
+				}
+				exit(1);
+			}
+			printf("New connection, peer addresses\n");
+			print_addr_buf(addrs, rc);
+			ra_family = addrs[0].sa_family;
+			switch (ra_family) {
+			case AF_INET:
+				ra_len = sizeof(remote_addr.v4);
+				break;
+			case AF_INET6:
+				ra_len = sizeof(remote_addr.v6);
+				break;
+			default:
+				fprintf(stderr, "Invalid address type.\n");
+				exit(1);
+			}
+			memcpy(&remote_addr, &addrs[0], ra_len);
+			sctp_freepaddrs(addrs);
+			new_connection = 0;
 		}
 
 		do {
-			error = sendmsg(sk, &outmsg, 0);
+			if (SOCK_SEQPACKET == socket_type ||
+			    (connectx_count == 0 && new_connection)) {
+				/* Initialize the message struct we use to pass
+				 * messages to the remote socket.
+				 */
+				if (!interactive_mode) {
+					iov.iov_base = message;
+					iov.iov_len = msglen;
+				}
+				outmsg.msg_iov = &iov;
+				outmsg.msg_iovlen = 1;
+				outmsg.msg_control = NULL;
+				outmsg.msg_controllen = 0;
+				outmsg.msg_name = &remote_addr;
+				outmsg.msg_namelen = ra_len;
+
+				error = sendmsg(sk, &outmsg, 0);
+			} else {
+				error = send(sk, message, msglen, 0);
+				if (error == -1 && errno == EPIPE) {
+					error = close(sk);
+					if (error != 0) {
+						fprintf(stderr, "close failed %s\n", strerror(errno));
+						exit(1);
+					}
+					*skp = sk = build_endpoint(argv0, local_port);
+					break;
+				}
+			}
 
 			if (error != msglen) {
 				fprintf(stderr, "%s: error: %s.\n",
@@ -830,8 +985,26 @@ command_send(char *argv0, int sk)
 			associd = test_recv_assoc_change(sk);
 
 		/* Verify there is no association.  */
-		if (0 != test_sk_for_assoc(sk, associd))
+		if (0 != test_sk_for_assoc(sk, associd)) {
 			printf("No association is present now!!\n");
+			new_connection = 1;
+		} else {
+			if (new_connection) {
+				int rc = sctp_getpaddrs(sk, associd, &addrs);
+				if (0 >= rc) {
+					if (rc == 0) {
+						fprintf(stderr, "sctp_getpaddrs failed, no peers.\n");
+					} else {
+						fprintf(stderr, "sctp_getpaddrs failed %s(%d).\n", strerror(errno), errno);
+					}
+					exit(1);
+				}
+				printf("New connection, peer addresses\n");
+				print_addr_buf(addrs, rc);
+				sctp_freepaddrs(addrs);
+				new_connection = 0;
+			}
+		}
 
 		/* Clean up.  */
 		if (interactive_mode) {
@@ -861,10 +1034,6 @@ command_poll(char *argv0)
 
 	struct msghdr outmsg;
 	struct hostent *hst;
-	sockaddr_storage_t remote_addr;
-	sa_family_t ra_family;	/* What family is remote_addr? */
-	int ra_len;		/* How long is remote_addr? */
-	void *ra_raw;		/* This is the addr part of remote_addr. */
 	int msglen;
 	int temp_fd, temp_set;
 
@@ -1158,13 +1327,19 @@ usage(char *argv0)
 		" -B, --bindx-add"
 		"\tadd the specified address(es) as additional bind\n"
 		"\t\t\taddresses to the local socket. Multiple addresses can\n"
-		"\t\t\tbe specified by this argument multiple times.\n"
+		"\t\t\tbe specified by using this argument multiple times.\n"
 		"\t\t\tFor example, '-B 10.0.0.1 -B 20.0.0.2'.\n"
 		" -b, --bindx-rem"
 		"\tremove the specified address(es) from the bind\n"
 		"\t\t\taddresses of the local socket. Multiple addresses can\n"
-		"\t\t\tbe specified by this argument multiple times.\n"
+		"\t\t\tbe specified by using this argument multiple times.\n"
 		"\t\t\tFor example, '-b 10.0.0.1 -b 20.0.0.2'.\n"
+		" -c, --connectx"
+		"\t\tuse the specified address(es) for connection to the\n"
+		"\t\t\tpeer socket. Multiple addresses can be specified by\n"
+		"\t\t\tusing this argument multiple times.\n"
+		"\t\t\tFor example, '-c 10.0.0.1 -c 20.0.0.2'.\n"
+		"\t\t\tThis option is incompatible with the -h option.\n"
 		" -I\t\t\tuse the interactive mode.\n"
 		" -i\t\t\tsetup the specified number of endpoints by using the\n"
 		"\t\t\tspecified local host (-H) and local port (-P). The port\n"
@@ -1413,12 +1588,14 @@ parse_inter_commands(char *argv0, char *input, int snd_only)
 			case INTER_BINDX_ADD:
 				tmp_addrs = get_bindx_addr(p, &len);
 				bindx_func(argv0, inter_sk, tmp_addrs, len,
-					SCTP_BINDX_ADD_ADDR);
+					SCTP_BINDX_ADD_ADDR, local_port);
+				free(tmp_addrs);
 				break;
 			case INTER_BINDX_REM:
 				tmp_addrs = get_bindx_addr(p, &len);
 				bindx_func(argv0, inter_sk, tmp_addrs, len,
-					SCTP_BINDX_REM_ADDR);
+					SCTP_BINDX_REM_ADDR, local_port);
+				free(tmp_addrs);
 				break;
 			case INTER_SET_PRIM:
 				primary_func(argv0, inter_sk, p, set);
@@ -1427,10 +1604,10 @@ parse_inter_commands(char *argv0, char *input, int snd_only)
 				peer_primary_func(argv0, inter_sk, p, set);
 				break;
 			case INTER_SHUTDOWN:
-				shutdown_func(argv0, inter_sk);
+				shutdown_func(argv0, &inter_sk, SHUTDOWN_SHUTDOWN);
 				break;
 			case INTER_ABORT:
-				abort_func(argv0, inter_sk);
+				shutdown_func(argv0, &inter_sk, SHUTDOWN_ABORT);
 				break;
 			case INTER_NODELAY:
 				if (set) {
@@ -1508,7 +1685,7 @@ static void
 sndbuf_func(char *argv0, int sk, int len, int set)
 {
 	int error;
-	int optlen;
+	socklen_t optlen;
 
 	if (set) {
 		error = setsockopt(sk, SOL_SOCKET, SO_SNDBUF,
@@ -1534,7 +1711,7 @@ static void
 rcvbuf_func(char *argv0, int sk, int len, int set)
 {
 	int error;
-	int optlen;
+	socklen_t optlen;
 
 	if (set) {
 		error = setsockopt(sk, SOL_SOCKET, SO_RCVBUF,
@@ -1587,7 +1764,7 @@ get_bindx_addr(char *in, int *count)
 } /* get_bindx_addr() */
 
 static int
-bindx_func(char *argv0, int sk, struct sockaddr *addrs, int count, int flag)
+bindx_func(char *argv0, int sk, struct sockaddr *addrs, int count, int flag, int portnum)
 {
 
 	int error;
@@ -1596,7 +1773,7 @@ bindx_func(char *argv0, int sk, struct sockaddr *addrs, int count, int flag)
 	void *aptr;
 
 
-	if (0 == local_port) {
+	if (0 == portnum) {
 		fprintf(stderr, "%s: A non-0 local port number is ", argv0);
 		fprintf(stderr, "required for bindx to work!\n");
 		return -1 ;
@@ -1610,12 +1787,12 @@ bindx_func(char *argv0, int sk, struct sockaddr *addrs, int count, int flag)
 		switch(sa_addr->sa_family) {
 		case AF_INET:
 			((struct sockaddr_in *)sa_addr)->sin_port =
-				htons(local_port);
+				htons(portnum);
 			aptr += sizeof(struct sockaddr_in);
 			break;
 		case AF_INET6:
 			((struct sockaddr_in6 *)sa_addr)->sin6_port =
-				htons(local_port);
+				htons(portnum);
 			aptr += sizeof(struct sockaddr_in6);
 			break;
 		default:
@@ -1625,7 +1802,6 @@ bindx_func(char *argv0, int sk, struct sockaddr *addrs, int count, int flag)
 	}
 
 	error = sctp_bindx(sk, addrs, count, flag);
-	free(addrs);
 
 	if (error != 0) {
 		if (flag == SCTP_BINDX_ADD_ADDR) {
@@ -1643,6 +1819,58 @@ bindx_func(char *argv0, int sk, struct sockaddr *addrs, int count, int flag)
 
 } /* bindx_func() */
 
+static int
+connectx_func(char *argv0, int sk, struct sockaddr *addrs, int count)
+{
+
+	int error;
+	int i;
+	struct sockaddr *sa_addr;
+	void *aptr;
+
+
+	if (0 == remote_port) {
+		fprintf(stderr, "%s: A non-0 remote port number is ", argv0);
+		fprintf(stderr, "required for connectx to work!\n");
+		return -1 ;
+	}
+
+	/* Set the port in every address.  */
+	aptr = addrs;
+	for (i = 0; i < count; i++) {
+		sa_addr = (struct sockaddr *)aptr;
+
+		switch(sa_addr->sa_family) {
+		case AF_INET:
+			((struct sockaddr_in *)sa_addr)->sin_port =
+				htons(remote_port);
+			aptr += sizeof(struct sockaddr_in);
+			break;
+		case AF_INET6:
+			((struct sockaddr_in6 *)sa_addr)->sin6_port =
+				htons(remote_port);
+			aptr += sizeof(struct sockaddr_in6);
+			break;
+		default:
+			fprintf(stderr, "Invalid address family\n");
+			return -1;
+		}
+	}
+
+	error = sctp_connectx(sk, addrs, count);
+
+	if (error != 0) {
+		if (errno == ECONNREFUSED)
+			return -2;
+		fprintf(stderr, "%s: error connecting to addrs: %s.\n",
+			argv0, strerror(errno));
+		return -1;
+	}
+
+	return 0;
+
+} /* connectx_func() */
+
 static void
 primary_func(char *argv0, int sk, char *cp, int set)
 {
@@ -1650,7 +1878,8 @@ primary_func(char *argv0, int sk, char *cp, int set)
 	struct sockaddr_in *in_addr;
 	struct sockaddr_in6 *in6_addr;
 	struct sockaddr *saddr;
-	int prim_len, ret;
+	socklen_t prim_len;
+	int ret;
 	char *p = cp;
 	char addr_buf[INET6_ADDRSTRLEN];
 	const char *ap = NULL;
@@ -1765,113 +1994,10 @@ err:
 	        (set)?"setting":"getting", strerror(errno));
 }
 
-static int
-shutdown_func(char *argv0, int sk)
-{
-	struct msghdr outmessage;
-	char outcmsg[CMSG_SPACE(sizeof(struct sctp_sndrcvinfo))];
-	struct cmsghdr *cmsg;
-        int error=0, bytes_sent;
-	struct sctp_sndrcvinfo *sinfo;
-	struct hostent *hst;
-
-	sockaddr_storage_t remote_addr;
-	sa_family_t ra_family;	/* What family is remote_addr? */
-	int ra_len;		/* How long is remote_addr? */
-	void *ra_raw;		/* This is the addr part of remote_addr. */
-
-	/* Set up the destination.  */
-	hst = gethostbyname(remote_host);
-	if (hst == NULL) {
-		hst = gethostbyname2(remote_host, AF_INET6);
-	}
-
-	if (hst == NULL || hst->h_length < 1) {
-		fprintf(stderr, "%s: bad hostname: %s\n",
-			argv0, remote_host);
-		exit(1);
-	}
-
-	ra_family = hst->h_addrtype;
-	switch (ra_family) {
-	case AF_INET:
-		ra_len = sizeof(remote_addr.v4);
-		ra_raw = &remote_addr.v4.sin_addr;
-		remote_addr.v4.sin_port = htons(remote_port);
-		remote_addr.v4.sin_family = AF_INET;
-		break;
-	case AF_INET6:
-		ra_len = sizeof(remote_addr.v6);
-		ra_raw = &remote_addr.v6.sin6_addr;
-		remote_addr.v6.sin6_port = htons(remote_port);
-		remote_addr.v6.sin6_family = AF_INET6;
-		break;
-	default:
-		fprintf(stderr, "Invalid address type.\n");
-		exit(1);
-		break;
-	}
-	memcpy(ra_raw, hst->h_addr_list[0], hst->h_length);
-
-
-	/* Initialize the message struct we use to pass messages to
-	 * the remote socket.
-	 */
-	outmessage.msg_name = &remote_addr;
-	outmessage.msg_namelen = ra_len;
-
-	outmessage.msg_iov = NULL;
-	outmessage.msg_iovlen = 0;
-	outmessage.msg_control = outcmsg;
-	outmessage.msg_controllen = sizeof(outcmsg);
-	outmessage.msg_flags = 0;
-
-	cmsg = CMSG_FIRSTHDR(&outmessage);
-	cmsg->cmsg_level = IPPROTO_SCTP;
-	cmsg->cmsg_type = SCTP_SNDRCV;
-	cmsg->cmsg_len = CMSG_LEN(sizeof(struct sctp_sndrcvinfo));
-	outmessage.msg_controllen = cmsg->cmsg_len;
-	sinfo = (struct sctp_sndrcvinfo *)CMSG_DATA(cmsg);
-	memset(sinfo, 0x00, sizeof(struct sctp_sndrcvinfo));
-	sinfo->sinfo_flags |= MSG_EOF;
-
-	sinfo->sinfo_assoc_id = associd;
-
-	/* Verify that the association is present. */
-	error = test_sk_for_assoc(sk, associd);
-	if (error != 0) {
-		printf("The association isn't present yet! "
-		       "Cannot shutdown!\n");
-		return -1;
-	}
-
-	bytes_sent = sendmsg(sk, &outmessage, 0);
-	if (bytes_sent != 0) {
-		printf("\tFailure:  %s.\n", strerror(errno));
-		return -1;
-	}
-
-	/* Receive the SHUTDOWN_COMP event for UDP-style sockets. */
-	if (socket_type == SOCK_SEQPACKET)
-		test_recv_assoc_change(sk);
-
-	/* Verify that the association is no longer present.  */
-	error = test_sk_for_assoc(sk, associd);
-	if (error != 0) {
-		printf("Successfully shudown the original association\n");
-		associd = 0;
-	} else {
-		printf("Shutdown failed\n");
-		exit(1);
-	}
-
-	return 0;
-}
-
 static int
 nodelay_func(char *argv0, int sk, int val, int set)
 {
-	int optlen;
+	socklen_t optlen;
 	int error;
 
 	if (set) {
@@ -1898,7 +2024,7 @@ nodelay_func(char *argv0, int sk, int val, int set)
 static int
 maxseg_func(char *argv0, int sk, int val, int set)
 {
-	int optlen;
+	socklen_t optlen;
 	int error;
 
 	if (set) {
@@ -1923,7 +2049,7 @@ maxseg_func(char *argv0, int sk, int val, int set)
 }
 
 static int
-abort_func(char *argv0, int sk)
+shutdown_func(char *argv0, int *skp, int shutdown_type)
 {
 	struct msghdr outmessage;
 	char outcmsg[CMSG_SPACE(sizeof(struct sctp_sndrcvinfo))];
@@ -1931,93 +2057,120 @@ abort_func(char *argv0, int sk)
         int error=0, bytes_sent;
 	struct sctp_sndrcvinfo *sinfo;
 	struct hostent *hst;
+	char *sd_type;
+	int sk = *skp;
 
-	sockaddr_storage_t remote_addr;
-	sa_family_t ra_family;	/* What family is remote_addr? */
-	int ra_len;		/* How long is remote_addr? */
-	void *ra_raw;		/* This is the addr part of remote_addr. */
-
-	/* Set up the destination.  */
-	hst = gethostbyname(remote_host);
-	if (hst == NULL) {
-		hst = gethostbyname2(remote_host, AF_INET6);
-	}
-
-	if (hst == NULL || hst->h_length < 1) {
-		fprintf(stderr, "%s: bad hostname: %s\n",
-			argv0, remote_host);
-		exit(1);
-	}
-
-	ra_family = hst->h_addrtype;
-	switch (ra_family) {
-	case AF_INET:
-		ra_len = sizeof(remote_addr.v4);
-		ra_raw = &remote_addr.v4.sin_addr;
-		remote_addr.v4.sin_port = htons(remote_port);
-		remote_addr.v4.sin_family = AF_INET;
-		break;
-	case AF_INET6:
-		ra_len = sizeof(remote_addr.v6);
-		ra_raw = &remote_addr.v6.sin6_addr;
-		remote_addr.v6.sin6_port = htons(remote_port);
-		remote_addr.v6.sin6_family = AF_INET6;
-		break;
-	default:
-		fprintf(stderr, "Invalid address type.\n");
-		exit(1);
-		break;
-	}
-	memcpy(ra_raw, hst->h_addr_list[0], hst->h_length);
-
-
-	/* Initialize the message struct we use to pass messages to
-	 * the remote socket.
-	 */
-	outmessage.msg_name = &remote_addr;
-	outmessage.msg_namelen = ra_len;
-
-	outmessage.msg_iov = NULL;
-	outmessage.msg_iovlen = 0;
-	outmessage.msg_control = outcmsg;
-	outmessage.msg_controllen = sizeof(outcmsg);
-	outmessage.msg_flags = 0;
-
-	cmsg = CMSG_FIRSTHDR(&outmessage);
-	cmsg->cmsg_level = IPPROTO_SCTP;
-	cmsg->cmsg_type = SCTP_SNDRCV;
-	cmsg->cmsg_len = CMSG_LEN(sizeof(struct sctp_sndrcvinfo));
-	outmessage.msg_controllen = cmsg->cmsg_len;
-	sinfo = (struct sctp_sndrcvinfo *)CMSG_DATA(cmsg);
-	memset(sinfo, 0x00, sizeof(struct sctp_sndrcvinfo));
-	sinfo->sinfo_flags |= MSG_ABORT;
-
-	sinfo->sinfo_assoc_id = associd;
+	if (shutdown_type == SHUTDOWN_ABORT)
+		sd_type = "ABORT";
+	else
+		sd_type = "SHUTDOWN";
 
 	/* Verify that the association is present. */
 	error = test_sk_for_assoc(sk, associd);
 	if (error != 0) {
-		printf("The association isn't present yet! Cannot abort!\n");
+		printf("The association isn't present yet! Cannot %s!\n", sd_type);
 		return -1;
 	}
 
-	bytes_sent = sendmsg(sk, &outmessage, 0);
-	if (bytes_sent != 0) {
-		printf("\tFailure:  %s.\n", strerror(errno));
-		return -1;
-	}
+	if (socket_type == SOCK_SEQPACKET) {
+		/* Set up the destination.  */
+		if (remote_host) {
+			hst = gethostbyname(remote_host);
+			if (hst == NULL) {
+				hst = gethostbyname2(remote_host, AF_INET6);
+			}
 
-	/* Receive the COMM_LOST event for UDP-style sockets. */
-	if (socket_type == SOCK_SEQPACKET)
+			if (hst == NULL || hst->h_length < 1) {
+				fprintf(stderr, "%s: bad hostname: %s\n",
+					argv0, remote_host);
+				exit(1);
+			}
+
+			ra_family = hst->h_addrtype;
+			switch (ra_family) {
+			case AF_INET:
+				ra_len = sizeof(remote_addr.v4);
+				ra_raw = &remote_addr.v4.sin_addr;
+				remote_addr.v4.sin_port = htons(remote_port);
+				remote_addr.v4.sin_family = AF_INET;
+				break;
+			case AF_INET6:
+				ra_len = sizeof(remote_addr.v6);
+				ra_raw = &remote_addr.v6.sin6_addr;
+				remote_addr.v6.sin6_port = htons(remote_port);
+				remote_addr.v6.sin6_family = AF_INET6;
+				break;
+			default:
+				fprintf(stderr, "Invalid address type.\n");
+				exit(1);
+				break;
+			}
+			memcpy(ra_raw, hst->h_addr_list[0], hst->h_length);
+		}
+
+		/* Initialize the message struct we use to pass messages to
+		 * the remote socket.
+		 */
+		outmessage.msg_name = &remote_addr;
+		outmessage.msg_namelen = ra_len;
+
+		outmessage.msg_iov = NULL;
+		outmessage.msg_iovlen = 0;
+		outmessage.msg_control = outcmsg;
+		outmessage.msg_controllen = sizeof(outcmsg);
+		outmessage.msg_flags = 0;
+
+		cmsg = CMSG_FIRSTHDR(&outmessage);
+		cmsg->cmsg_level = IPPROTO_SCTP;
+		cmsg->cmsg_type = SCTP_SNDRCV;
+		cmsg->cmsg_len = CMSG_LEN(sizeof(struct sctp_sndrcvinfo));
+		outmessage.msg_controllen = cmsg->cmsg_len;
+		sinfo = (struct sctp_sndrcvinfo *)CMSG_DATA(cmsg);
+		memset(sinfo, 0x00, sizeof(struct sctp_sndrcvinfo));
+		if (shutdown_type == SHUTDOWN_ABORT)
+			sinfo->sinfo_flags |= SCTP_ABORT;
+		else
+			sinfo->sinfo_flags |= SCTP_EOF;
+
+		sinfo->sinfo_assoc_id = associd;
+	
+		bytes_sent = sendmsg(sk, &outmessage, 0);
+		if (bytes_sent != 0) {
+			printf("Failure:  %s.\n", strerror(errno));
+			return -1;
+		}
+
+		/* Receive the COMM_LOST or SHUTDOWN_COMP event. */
 		test_recv_assoc_change(sk);
+	} else {
+		if (shutdown_type == SHUTDOWN_ABORT) {
+			struct  linger {
+				int  l_onoff; 
+				int  l_linger;
+			} data = {1, 0};
+			error = setsockopt(sk, SOL_SOCKET, SO_LINGER,
+					   (char *)&data, sizeof(data));
+			if (error != 0) {
+				printf("setsockopt failed %s\n", strerror(errno));
+				exit(1);
+			}
+		}
+		error = close(sk);
+		if (error != 0) {
+			printf("close failed %s\n", strerror(errno));
+			exit(1);
+		}
+		*skp = sk = build_endpoint(argv0, local_port);
+	}
 
 	/* Verify that the association is no longer present.  */
 	error = test_sk_for_assoc(sk, associd);
 	if (error != 0) {
-		printf("Successfully abort the original association\n");
+		printf("Successfully %s the original association\n", sd_type);
 		associd = 0;
+		new_connection = 1;
 	} else {
-		printf("Abort failed\n");
+		printf("%s failed\n", sd_type);
 		exit(1);
 	}
 
@@ -2029,7 +2182,7 @@ test_sk_for_assoc(int sk, sctp_assoc_t assoc_id)
 {
 	int error = 0;
 	struct sctp_status status;
-	int status_len;
+	socklen_t status_len;
 
 	memset(&status, 0, sizeof(status));
 	if (assoc_id)
@@ -2082,20 +2235,70 @@ test_verify_assoc_change(struct msghdr *msg)
 	union sctp_notification *sn;
 
 	if (!(msg->msg_flags & MSG_NOTIFICATION)) {
-		printf("%s: Received data when notification is expected\n",
+		fprintf(stderr, "%s: Received data when notification is expected\n",
 		       __FUNCTION__);
 		exit(1);
 	}
 
 	sn = (union sctp_notification *)msg->msg_iov->iov_base;
 	if (SCTP_ASSOC_CHANGE != sn->sn_header.sn_type) {
-		printf("%s: Received unexpected notification: %d",
+		fprintf(stderr, "%s: Received unexpected notification: %d",
 			__FUNCTION__, sn->sn_header.sn_type);
 		exit(1);
 	}
 
+	switch(sn->sn_assoc_change.sac_state)
+	{
+	case SCTP_COMM_UP:
+		printf("Recieved SCTP_COMM_UP\n");
+		break;
+	case SCTP_COMM_LOST:
+		printf("Recieved SCTP_COMM_LOST\n");
+		break;
+	case SCTP_RESTART:
+		printf("Recieved SCTP_RESTART\n");
+		break;
+	case SCTP_SHUTDOWN_COMP:
+		printf("Recieved SCTP_SHUTDOWN_COMP\n");
+		break;
+	case SCTP_CANT_STR_ASSOC:
+		printf("Recieved SCTP_CANT_STR_ASSOC\n");
+		break;
+	}
+
 	if (SCTP_COMM_UP == sn->sn_assoc_change.sac_state)
 		return sn->sn_assoc_change.sac_assoc_id;
 	else
 		return 0;
 }
+
+void print_addr_buf(void * laddrs, int n_laddrs)
+{
+	void *addr_buf = laddrs;
+	int i;
+
+	for (i = 0; i < n_laddrs; i++) {
+		addr_buf += print_sockaddr((struct sockaddr *)addr_buf);
+		printf("\n");
+	}
+}
+
+int print_sockaddr(struct sockaddr *sa_addr)
+{
+	struct sockaddr_in *in_addr;
+	struct sockaddr_in6 *in6_addr;
+
+	if (AF_INET == sa_addr->sa_family) {
+		in_addr = (struct sockaddr_in *)sa_addr;
+		printf("%d.%d.%d.%d:%d",
+		       NIPQUAD(in_addr->sin_addr),
+		       ntohs(in_addr->sin_port));
+		return sizeof(struct sockaddr_in);
+	} else {
+		in6_addr = (struct sockaddr_in6 *)sa_addr;
+		printf("%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x:%d",
+		       NIP6(in6_addr->sin6_addr),
+		       ntohs(in6_addr->sin6_port));
+		return sizeof(struct sockaddr_in6);
+	}
+}
