@@ -87,11 +87,6 @@ extern struct list_head test_timers;
 
 int ft_sctp_lock_bug = 0;
 int ft_sctp_lock_assert = 0;
-#ifdef JUNK
-struct timeval xtime = {0,0};
-#else
-struct timeval ytime = {0,0};
-#endif
 unsigned long volatile jiffies = 0;
 
 cpumask_t cpu_callout_map;
@@ -2726,7 +2721,7 @@ int __xfrm_policy_check(struct sock *sk, int dir, struct sk_buff *skb, unsigned
 }
 
 #ifdef CONFIG_SMP
-void *__alloc_percpu(size_t size, size_t align)
+void *__alloc_percpu(size_t size)
 {
 	        return(kmalloc(size, GFP_ATOMIC));
 }
@@ -2737,7 +2732,7 @@ void free_percpu(const void *ptr)
 #endif
 
 struct sock_filter;
-int sk_run_filter(struct sk_buff *skb, struct sock_filter *filter, int flen)
+unsigned int sk_run_filter(struct sk_buff *skb, struct sock_filter *filter, int flen)
 {
 	return 0;
 }
@@ -3208,3 +3203,8 @@ int find_next_bit(const unsigned long *addr, int size, int offset)
 
 	return offset;
 }
+
+int capable(int cap)
+{
+	return 1;
+}
